#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import traceback
import base64
import json
import datetime

from langchain.schema import Document

from ..components import get_embeddings, pick_vectorstore, load_memories, pick_embedding
from ..logging import setup_logging

logging = setup_logging()

def embed_pubsub_chunk(data: dict):
    """Triggered from a message on a Cloud Pub/Sub topic "embed_chunk" topic
    Will only attempt to send one chunk to vectorstore.
    Args:
         data JSON
    """

    message_data = base64.b64decode(data['message']['data']).decode('utf-8')
    messageId = data['message'].get('messageId')
    publishTime = data['message'].get('publishTime')

    logging.debug(f"This Function was triggered by messageId {messageId} published at {publishTime}")
    logging.debug(f"from_pubsub_to_supabase message data: {message_data}")

    try:
        the_json = json.loads(message_data)
    except Exception as err:
        logging.error(f"Error - could not parse message_data: {err}: {message_data}")
        return "Could not parse message_data"

    if not isinstance(the_json, dict):
        raise ValueError(f"Could not parse message_data from json to a dict: got {message_data} or type: {type(the_json)}")

    page_content = the_json.get("page_content", None)
    if page_content is None:
        return "No page content"
    if len(page_content) < 100:
        logging.warning(f"too little page content to add: {message_data}")
        return "Too little characters"
    
    metadata = the_json.get("metadata", None)

    vector_name = metadata.get("vector_name", None)
    if vector_name is None:
        msg = f"FATAL: No vector name was found within metadata: {metadata}"
        logging.error(msg)
        return msg
    
    logging.info(f"Embedding: {vector_name} page_content: {page_content[:30]}...")

    if 'eventTime' not in metadata:
        metadata['eventTime'] = datetime.datetime.utcnow().isoformat(timespec='microseconds') + "Z"
    metadata['eventtime'] = metadata['eventTime'] # case insensitive bugs

    if 'source' not in metadata:
        if 'objectId' in metadata:
            metadata['source'] = metadata['objectId']
    
    metadata['original_source'] = metadata['source']

    # add metadata to page_content too for vectorstores that don't suppoort the metdata field
    chunk_metadata = f"## Chunk Metadata:\n{json.dumps(metadata)}\n"
    page_content = f"## Chunk Content:\n{page_content}\n{chunk_metadata}"

    doc = Document(page_content=page_content, metadata=metadata)

    # init embedding and vector store
    embeddings = get_embeddings(vector_name)

    memories = load_memories(vector_name)
    vectorstore_list = []
    for memory in memories:  # Iterate over the list
        for key, value in memory.items(): 
            logging.info(f"Found memory {key}")
            vectorstore = value.get('vectorstore', None)
            if vectorstore is not None:
                # check if vectorstore specific embedding is available
                embed_llm = value.get('llm', None)
                if embed_llm is not None:
                    embeddings = pick_embedding(embed_llm)
                vectorstore_obj = pick_vectorstore(vectorstore, vector_name=vector_name, embeddings=embeddings)
                vs_retriever = vectorstore_obj.as_retriever(search_kwargs=dict(k=3))
                vectorstore_list.append(vs_retriever)

    # can have multiple vectorstores per embed
    metadata_list = []
    import uuid
    for vector_store in vectorstore_list:
        logging.debug(f"Adding single document for {vector_name} to vector store {vector_store}")
        try:
            vector_store.add_documents([doc], ids = [str(uuid.uuid4())])
            logging.info(f"Added doc for {vector_name} to {vector_store} - metadata: {metadata}")
            metadata_list.append(metadata)
        except Exception as err:
            error_message = traceback.format_exc()
            logging.error(f"Could not add document {doc} for {vector_name} to {vector_store} - metadata: {metadata}: {str(err)} traceback: {error_message}")

    return metadata_list



#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import os
from ..logging import setup_logging
from ..utils.config import load_config_key

logging = setup_logging()


def pick_vectorstore(vs_str, vector_name, embeddings):
    logging.debug('Picking vectorstore')
        
    if vs_str == 'supabase':
        from supabase import Client, create_client
        from langchain.vectorstores import SupabaseVectorStore
        from ..database.database import setup_supabase

        logging.debug(f"Initiating Supabase store: {vector_name}")
        setup_supabase(vector_name)

        # init embedding and vector store
        supabase_url = os.getenv('SUPABASE_URL')
        supabase_key = os.getenv('SUPABASE_KEY')

        logging.debug(f"Supabase URL: {supabase_url} vector_name: {vector_name}")

        supabase: Client = create_client(supabase_url, supabase_key)

        vectorstore = SupabaseVectorStore(supabase, 
                                        embeddings,
                                        table_name=vector_name,
                                        query_name=f'match_documents_{vector_name}')

        logging.debug("Chose Supabase")

        return vectorstore
    
    elif vs_str == 'cloudsql':
        from langchain.vectorstores.pgvector import PGVector

        logging.debug("Inititaing CloudSQL pgvector")
        #setup_cloudsql(vector_name) 

        # https://python.langchain.com/docs/modules/data_connection/vectorstores/integrations/pgvector
        CONNECTION_STRING = os.environ.get("PGVECTOR_CONNECTION_STRING")
        # postgresql://brainuser:password@10.24.0.3:5432/brain

        from ..database.database import get_vector_size
        vector_size = get_vector_size(vector_name)

        os.environ["PGVECTOR_VECTOR_SIZE"] = str(vector_size)
        vectorstore = PGVector(connection_string=CONNECTION_STRING,
            embedding_function=embeddings,
            collection_name=vector_name,
            #pre_delete_collection=True # for testing purposes
            )

        logging.debug("Chose CloudSQL")

        return vectorstore
    
    elif vs_str == 'alloydb':
        from langchain_google_alloydb_pg import AlloyDBEngine, AlloyDBVectorStore
        from google.cloud.alloydb.connector import IPTypes
        from ..database.alloydb import create_alloydb_table

        alloydb_config = load_config_key(
            'alloydb_config', 
            vector_name=vector_name, 
            filename = "config/llm_config.yaml"
        )

        if alloydb_config is None:
            logging.error("No alloydb_config was found")

        ALLOYDB_DB = os.environ.get("ALLOYDB_DB")
        if ALLOYDB_DB is None:
            logging.error(f"Could not locate ALLOYDB_DB environment variable for {vector_name}")
        logging.info(f"ALLOYDB_DB environment variable found for {vector_name} - {ALLOYDB_DB}")
        
        logging.info("Inititaing AlloyDB Langchain")

        
        engine = AlloyDBEngine.from_instance(
            project_id=alloydb_config["project_id"],
            region=alloydb_config["region"],
            cluster=alloydb_config["cluster"],
            instance=alloydb_config["instance"],
            database=alloydb_config.get("database") or ALLOYDB_DB,
            ip_type=alloydb_config.get("ip_type") or IPTypes.PRIVATE
        )

        create_alloydb_table(vector_name, engine)

        logging.info("Chose AlloyDB")
        vectorstore = AlloyDBVectorStore.create_sync(
                engine=engine,
                table_name=vector_name,
                embedding_service=embeddings,
                metadata_columns=["source"]
                #metadata_columns=["source", "eventTime"]
            )
        return vectorstore
        
    elif vs_str == "lancedb":
        from ..patches.langchain.lancedb import LanceDB
        import lancedb

        LANCEDB_BUCKET = os.environ.get("LANCEDB_BUCKET")
        if LANCEDB_BUCKET is None:
            logging.error(f"Could not locate LANCEDB_BUCKET environment variable for {vector_name}")
        logging.info(f"LANCEDB_BUCKET environment variable found for {vector_name} - {LANCEDB_BUCKET}")

        db = lancedb.connect(LANCEDB_BUCKET)

        logging.info(f"LanceDB Tables: {db.table_names()} using {LANCEDB_BUCKET}")
        logging.info(f"Opening LanceDB table: {vector_name} using {LANCEDB_BUCKET}")
    
        try:
            table = db.open_table(vector_name)
        except FileNotFoundError as err:
            logging.info(f"{err} - Could not open table for {vector_name} - creating new table")
            init = f"Creating new table for {vector_name}"
            table = db.create_table(
                        vector_name,
                        data=[
                            {
                                "vector": embeddings.embed_query(init),
                                "text": init,
                                "id": "1",
                            }
                        ],
                        mode="overwrite",
                    )

        logging.info(f"Inititaing LanceDB object for {vector_name} using {LANCEDB_BUCKET}")
        vectorstore = LanceDB(
            connection=table,
            embedding=embeddings,
        )
        logging.info(f"Chose LanceDB for {vector_name} using {LANCEDB_BUCKET}")

        return vectorstore

    else:
        raise NotImplementedError(f'No llm implemented for {vs_str}')   



#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from ..logging import setup_logging

import pathlib
import tempfile
import os
import re
import json

from google.cloud import storage
from langchain.schema import Document

from .gcs import add_file_to_gcs
from .splitter import chunk_doc_to_docs
from .pdfs import split_pdf_to_pages
from .publish import publish_if_urls
from . import loaders
from ..utils.parsers import extract_urls

logging = setup_logging()

def handle_gcs_message(message_data: str, metadata: dict, vector_name: str):
    # Process message from Google Cloud Storage
    logging.debug("Detected gs://")
    bucket_name, file_name = message_data[5:].split("/", 1)

    # Create a client
    storage_client = storage.Client()

    # Download the file from GCS
    bucket = storage_client.get_bucket(bucket_name)
    blob = bucket.blob(file_name)

    file_name=pathlib.Path(file_name)

    with tempfile.TemporaryDirectory() as temp_dir:
        tmp_file_path = os.path.join(temp_dir, file_name.name)
        blob.download_to_filename(tmp_file_path)

        if file_name.suffix == ".pdf":
            pages = split_pdf_to_pages(tmp_file_path, temp_dir)
            if len(pages) > 1: # we send it back to GCS to parrallise the imports
                logging.info(f"Got back {len(pages)} pages for file {tmp_file_path}")
                for pp in pages:
                    gs_file = add_file_to_gcs(pp, vector_name=vector_name, bucket_name=bucket_name, metadata=metadata)
                    logging.info(f"{gs_file} is now in bucket {bucket_name}")
                logging.info(f"Sent split pages for {file_name.name} back to GCS to parrallise the imports")
                return None
        else:
            # just original temp file
            pages = [tmp_file_path]

        the_metadata = {
            "source": message_data,
            "type": "file_load_gcs",
            "bucket_name": bucket_name
        }
        metadata.update(the_metadata)

        docs = []
        for page in pages:
            logging.info(f"Sending file {page} to loaders.read_file_to_document {metadata}")
            docs2 = loaders.read_file_to_document(page, metadata=metadata)
            docs.extend(docs2)

        chunks = chunk_doc_to_docs(docs, file_name.suffix, vector_name=vector_name)

        return chunks, metadata

def handle_google_drive_message(message_data: str, metadata: dict, vector_name: str):
    # Process message from Google Drive
    logging.info("Got google drive URL")
    urls = extract_urls(message_data)

    docs = []
    for url in urls:
        metadata["source"] = url
        metadata["url"] = url
        metadata["type"] = "url_load"
        doc = loaders.read_gdrive_to_document(url, metadata=metadata)
        if doc is None:
            logging.info("Could not load any Google Drive docs")
        else:
            docs.extend(doc)

    chunks = chunk_doc_to_docs(docs, vector_name=vector_name)

    return chunks, metadata

def handle_github_message(message_data: str, metadata: dict, vector_name: str):
    # Process message from GitHub
    logging.info("Got GitHub URL")
    urls = extract_urls(message_data)

    branch="main"
    if "branch:" in message_data:
        match = re.search(r'branch:(\w+)', message_data)
        if match:
            branch = match.group(1)
    
    logging.info(f"Using branch: {branch}")

    docs = []
    for url in urls:
        metadata["source"] = url
        metadata["url"] = url
        metadata["type"] = "url_load"
        doc = loaders.read_git_repo(url, branch=branch, metadata=metadata)
        if doc is None:
            logging.info("Could not load GitHub files")
        else:
            docs.extend(doc)
    
    chunks = chunk_doc_to_docs(docs, vector_name=vector_name)
    
    return chunks, metadata

def handle_http_message(message_data: str, metadata: dict, vector_name:str):
    # Process message from a generic HTTP URL
    logging.info(f"Got http message: {message_data}")

    # just in case, extract the URL again
    urls = extract_urls(message_data)

    docs = []
    for url in urls:
        metadata["source"] = url
        metadata["url"] = url
        metadata["type"] = "url_load"
        doc = loaders.read_url_to_document(url, metadata=metadata)
        docs.extend(doc)

    chunks = chunk_doc_to_docs(docs, vector_name=vector_name)

    return chunks, metadata

def handle_json_content_message(message_data: str, metadata: dict, vector_name: str):
    logging.info("No tailored message_data detected, processing message json")
    # Process message containing direct JSON content
    the_json = json.loads(message_data)
    the_metadata = the_json.get("metadata", {})
    metadata.update(the_metadata)
    the_content = the_json.get("page_content", None)

    if metadata.get("source", None) is not None:
        metadata["source"] = "No source embedded"

    if the_content is None:
        logging.info("No content found")
        return {"metadata": "No content found"}
    
    docs = [Document(page_content=the_content, metadata=metadata)]

    publish_if_urls(the_content, vector_name)

    chunks = chunk_doc_to_docs(docs,vector_name=vector_name)

    return chunks, metadata


#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import time
from httpcore import ReadTimeout
from httpx import ReadTimeout
import traceback
from ..logging import setup_logging

logging = setup_logging()

def retry_qna(qa_function, question, max_retries=1, initial_delay=5):
    for retry in range(max_retries):
        try:
            return qa_function(question)
        except ReadTimeout as err:
            delay = initial_delay * (retry + 1)
            logging.warning(f"Read timeout while asking: {question} - trying again after {delay} seconds. Error: {str(err)}")
            time.sleep(delay)
            try:
                result = qa_function(question)
                result["answer"] = result["answer"] + " (Sorry for the delay, brain was a bit slow - should be quicker next time)"
                return result
            except ReadTimeout:
                if retry == max_retries - 1:
                    raise
        except Exception as err:
            delay = initial_delay * (retry + 1)
            logging.error(f"General error: {traceback.format_exc()}")
            time.sleep(delay)
            try:
                result = qa_function(question)
                result["answer"] = result["answer"] + " (Sorry for the delay, had to warm up the brain - should be quicker next time)"
                return result
            except Exception:
                if retry == max_retries - 1:
                    raise

    raise Exception(f"Max retries exceeded for question: {question}")


import pytest
import requests_mock
from sunholo.agents import dispatch_to_qa


def setup():
    user_input = 'mock_user_input'
    chat_history = 'mock_chat_history'
    vector_name = 'mock_vector_name'
    stream = 'mock_stream'
    return user_input, chat_history, vector_name, stream


def test_prep_request_payload():
    user_input, chat_history, vector_name, stream = setup()
    # Call 'prep_request_payload' with the mock data
    qna_endpoint, qna_data = dispatch_to_qa.prep_request_payload(user_input, chat_history, vector_name, stream)
    # Check the returned 'qna_endpoint' and 'qna_data'
    assert qna_endpoint == 'expected_qna_endpoint'
    assert qna_data == 'expected_qna_data'


def test_send_to_qa():
    user_input, chat_history, vector_name, stream = setup()
    mock_qna_endpoint = 'http://mock.qna.endpoint'
    mock_response = {'status': 'success'}
    with requests_mock.Mocker() as m:
        m.post(mock_qna_endpoint, json=mock_response)
        response = dispatch_to_qa.send_to_qa(user_input, chat_history, vector_name, stream)
        assert response == mock_response

    # Scenario where 'stream' is True and the response is a generator that yields response content chunks
    mock_response_chunk = 'mock_response_chunk'
    m.post(mock_qna_endpoint, text=mock_response_chunk)
    response = dispatch_to_qa.send_to_qa(user_input, chat_history, vector_name, True)
    assert next(response) == mock_response_chunk

    # Scenario where an HTTP error occurs
    m.post(mock_qna_endpoint, status_code=500)
    response = dispatch_to_qa.send_to_qa(user_input, chat_history, vector_name, stream)
    assert 'Error' in response

    # Scenario where any other error occurs
    m.post(mock_qna_endpoint, exc=requests_mock.exceptions.ReadTimeout)
    response = dispatch_to_qa.send_to_qa(user_input, chat_history, vector_name, stream)
    assert 'Error' in response


def test_send_to_qa_async():
    user_input, chat_history, vector_name, stream = setup()
    mock_qna_endpoint = 'http://mock.qna.endpoint'
    mock_response = {'status': 'success'}
    import aiohttp

    @pytest.mark.asyncio
    async def test_send_to_qa_async():
        user_input, chat_history, vector_name, stream = setup()
        mock_qna_endpoint = 'http://mock.qna.endpoint'
        mock_response = {'status': 'success'}
        async with aiohttp.ClientSession() as session:
            async with session.post(mock_qna_endpoint, json=mock_response) as resp:
                response = await dispatch_to_qa.send_to_qa_async(user_input, chat_history, vector_name, stream)
                assert response == mock_response

        # Scenario where an HTTP error occurs
        async with session.post(mock_qna_endpoint, status=500) as resp:
            response = await dispatch_to_qa.send_to_qa_async(user_input, chat_history, vector_name, stream)
            assert 'Error' in response

        # Scenario where any other error occurs
        async with session.post(mock_qna_endpoint, raise_for_status=False) as resp:
            response = await dispatch_to_qa.send_to_qa_async(user_input, chat_history, vector_name, stream)
            assert 'Error' in response


#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from flask import request, jsonify, Response

import json
import traceback

from ...agents import extract_chat_history, handle_special_commands
from ...qna.parsers import parse_output
from ...streaming import start_streaming_chat
from ...archive import archive_qa
from ...logging import setup_logging

logging = setup_logging()

def register_qna_routes(app, stream_interpreter, qna_interpreter):
    @app.route('/qna/streaming/<vector_name>', methods=['POST'])
    def stream_qa(vector_name):
        data = request.get_json()

        user_input = data['user_input'].strip()  # Extract user input from the payload
        chat_history = data.get('chat_history', None)
        stream_wait_time = data.get('stream_wait_time', 7)
        stream_timeout = data.get('stream_timeout', 120)

        # kwargs for streaming extracted from the payload here
        message_author = data.get('message_author', None)

        paired_messages = extract_chat_history(chat_history)

        command_response = handle_special_commands(user_input, vector_name, paired_messages)
        if command_response is not None:
            return jsonify(command_response)

        logging.info(f'Streaming data with stream_wait_time: {stream_wait_time} and stream_timeout: {stream_timeout}')
        def generate_response_content():
            for chunk in start_streaming_chat(user_input,
                                              vector_name=vector_name,
                                              qna_func=stream_interpreter,
                                              chat_history=paired_messages,
                                              wait_time=stream_wait_time,
                                              timeout=stream_timeout,
                                              #kwargs
                                              message_author=message_author
                                              ):
                if isinstance(chunk, dict) and 'answer' in chunk:
                    # When we encounter the dictionary, we yield it as a JSON string
                    # and stop the generator.
                    archive_qa(chunk, vector_name)
                    yield f"###JSON_START###{json.dumps(chunk)}###JSON_END###"
                    return
                else:
                    # Otherwise, we yield the plain text chunks as they come in.
                    yield chunk

        # Here, the generator function will handle streaming the content to the client.
        response = Response(generate_response_content(), content_type='text/plain; charset=utf-8')
        response.headers['Transfer-Encoding'] = 'chunked'    

        return response

    @app.route('/qna/<vector_name>', methods=['POST'])
    def process_qna(vector_name):
        # The body of the route handler goes here
        # Use `qna_interpreter` where the original code used `interpreter_qna`
        data = request.get_json()
        logging.info(f"qna/{vector_name} got data: {data}")

        user_input = data['user_input'].strip()

        message_author = data.get('message_author', None)

        paired_messages = extract_chat_history(data.get('chat_history', None))

        command_response = handle_special_commands(user_input, vector_name, paired_messages)
        if command_response is not None:
            return jsonify(command_response)

        try:
            bot_output = qna_interpreter(user_input, vector_name, chat_history=paired_messages, message_author=message_author)
            # {"answer": "The answer", "source_documents": [{"page_content": "The page content", "metadata": "The metadata"}]}
            bot_output = parse_output(bot_output)
            archive_qa(bot_output, vector_name)
        except Exception as err: 
            bot_output = {'answer': f'QNA_ERROR: An error occurred while processing /qna/{vector_name}: {str(err)} traceback: {traceback.format_exc()}'}
        
        logging.info(f'==LLM Q:{user_input} - A:{bot_output["answer"]}')
        
        return jsonify(bot_output)

    # Any other QNA related routes can be added here



#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
import re
import hashlib

def contains_url(message_data):
    url_pattern = re.compile(r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+')
    if url_pattern.search(message_data):
        return True
    else:
        return False

def extract_urls(text):
    url_pattern = re.compile(r'http[s]?://(?:[a-zA-Z]|[0-9]|[$-_@.&+]|[!*\\(\\),]|(?:%[0-9a-fA-F][0-9a-fA-F]))+')
    urls = url_pattern.findall(text)
    return urls

def compute_sha1_from_file(file_path):
    with open(file_path, "rb") as file:
        bytes = file.read() 
        readable_hash = hashlib.sha1(bytes).hexdigest()
    return readable_hash

def compute_sha1_from_content(content):
    readable_hash = hashlib.sha1(content).hexdigest()
    return readable_hash


def remove_whitespace(page_content: str):
    return page_content.replace("\n", " ").replace("\r", " ").replace("\t", " ").replace("  ", " ")


#   Copyright [2024] [Holosun ApS]
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
from ..logging import setup_logging
from ..utils.config import load_config_key, load_config, get_module_filepath

logging = setup_logging()

def pick_llm(vector_name):
    logging.debug('Picking llm')
    
    llm_str = load_config_key("llm", vector_name, filename = "config/llm_config.yaml")
    
    if llm_str == 'openai':
        llm_chat = get_llm_chat(vector_name)
        llm = get_llm_chat(vector_name, model="gpt-3.5-turbo-16k") # TODO: fix it needs llm_chat and not llm
        embeddings = get_embeddings(vector_name)
        logging.debug("Chose OpenAI")
    elif llm_str == 'vertex':
        llm = get_llm_chat(vector_name) # TODO: fix it needs llm_chat and not llm
        llm_chat = get_llm_chat(vector_name)
        embeddings = get_embeddings(vector_name)
        logging.debug("Chose VertexAI text-bison")
    elif llm_str == 'codey':
        llm = get_llm(vector_name)
        llm_chat = get_llm_chat(vector_name)
        embeddings = get_embeddings(vector_name)
        logging.debug("Chose VertexAI code-bison")
    elif llm_str == 'model_garden':
        llm = get_llm(vector_name)
        llm_chat = llm
        embeddings = None
        logging.debug("Chose VertexAIModelGarden")

    else:
        raise NotImplementedError(f'No llm implemented for {llm_str}')   

    return llm, embeddings, llm_chat

def pick_streaming(vector_name):
    
    llm_str = load_config_key("llm", vector_name, filename = "config/llm_config.yaml")
    
    if llm_str == 'openai' or llm_str == 'gemini' or llm_str == 'vertex':
        return True
    
    return False
 

def get_llm(vector_name, model=None, config_file="config/llm_config.yaml"):
    llm_str = load_config_key("llm", vector_name, filename=config_file)
    model_lookup_filepath = get_module_filepath("lookup/model_lookup.yaml")
    model_lookup, _ = load_config(model_lookup_filepath)

    if not model:
        model = load_config_key("model", vector_name, filename=config_file)

    logging.debug(f"Chose LLM: {llm_str}")
    # Configure LLMs based on llm_str
    if llm_str == 'openai':
        # Setup for OpenAI LLM
        from langchain_openai import ChatOpenAI
        if model is None:
            model = 'gpt-3.5-turbo'
            logging.info(f"No 'model' value in config file - selecting default ChatOpenAI: {model}")
            return ChatOpenAI(model=model, temperature=0, max_tokens=4000)
                
        return ChatOpenAI(model=model, temperature=0, max_tokens=4000)

    elif llm_str == 'vertex':
        # Setup for Vertex LLM
        from langchain_community.llms import VertexAI
        if model is None:
            model = 'text-unicorn'
            logging.info(f"No 'model' value in config file - selecting default {model}")
            
        return VertexAI(model_name = model, temperature=0, max_output_tokens=1024)

    elif llm_str == 'model_garden':
        from ..patches.langchain.vertexai import VertexAIModelGarden
        model_garden_config = load_config_key("gcp_config", vector_name, filename = config_file)
        if model_garden_config is None:
            raise ValueError("llm='model_garden' requires a gcp_config entry in config yaml file")
        
        return VertexAIModelGarden(project=model_garden_config['project_id'], 
                                   endpoint_id=model_garden_config['endpoint_id'], 
                                   location=model_garden_config['location'], 
                                   allowed_model_args=["max_tokens"])
    elif llm_str == 'anthropic':
        from langchain_anthropic import AnthropicLLM
        if model is None:
            model = 'claude-2.1'
            logging.info(f"No 'model' value in config file - selecting default {model}")
        return AnthropicLLM(model_name = model, temperature=0)

    if llm_str is None:
        raise NotImplementedError(f'No llm implemented for {llm_str}')

def get_llm_chat(vector_name, model=None, config_file="config/llm_config.yaml"):
    llm_str = load_config_key("llm", vector_name, filename=config_file)
    if not model:
        model = load_config_key("model", vector_name, filename=config_file)

    logging.debug(f"Chose LLM: {llm_str}")
    # Configure LLMs based on llm_str
    if llm_str == 'openai':
        # Setup for OpenAI LLM
        from langchain_openai import ChatOpenAI
        if model is None:
            model = 'gpt-4'
            logging.info(f"No 'model' value in config file - selecting default {model}")
            
        return ChatOpenAI(model=model, temperature=0, max_tokens=4000)

    elif llm_str == 'vertex':
        # Setup for Vertex LLM
        from langchain_community.chat_models import ChatVertexAI
        if model is None:
            model = 'gemini-1.0-pro'
            logging.info(f"No 'model' value in config file - selecting default {model}")
            
        return ChatVertexAI(model_name = model, temperature=0, max_output_tokens=1024)
    
    elif llm_str == 'gemini':
        from langchain_google_genai import ChatGoogleGenerativeAI
        if model is None:
            model="gemini-pro"
            logging.info(f"No 'model' value in config file - selecting default {model}")
        
        return ChatGoogleGenerativeAI(model_name = model, temperature=0)

    elif llm_str == 'anthropic':
        from langchain_anthropic import ChatAnthropic
        if model is None:
            model = 'claude-3-opus-20240229'
            logging.info(f"No 'model' value in config file - selecting default {model}")

        return ChatAnthropic(model_name = model, temperature=0)

    if llm_str is None:
        raise NotImplementedError(f'No llm implemented for {llm_str}')

def get_embeddings(vector_name):
    llm_str = load_config_key("llm", vector_name, filename="config/llm_config.yaml")

    return pick_embedding(llm_str)



def pick_embedding(llm_str):
    # get embedding directly from llm_str
    # Configure embeddings based on llm_str
    if llm_str == 'openai':
        # Setup for OpenAI embeddings
        from langchain_openai import OpenAIEmbeddings
        return OpenAIEmbeddings()
    elif llm_str == 'vertex' or llm_str == 'codey':
        # Setup for Text-Bison embeddings
        from langchain_community.embeddings import VertexAIEmbeddings
        
        return VertexAIEmbeddings()
    elif llm_str == 'gemini':
        from langchain_google_genai import GoogleGenerativeAIEmbeddings

        return GoogleGenerativeAIEmbeddings(model="models/embedding-001") #TODO add embedding type

    if llm_str is None:
        raise NotImplementedError(f'No embeddings implemented for {llm_str}')
